@@ -33,7 +33,7 @@
 /* This is a "major" version in the sense that it's only bumped
  * for incompatible changes.
  */
-#define SUPPORTED_GIR_VERSION "1.1"
+#define SUPPORTED_GIR_VERSION "1.2"
 
 struct _GIrParser
 {
@@ -47,33 +47,34 @@ typedef enum
   STATE_END,
   STATE_REPOSITORY,
   STATE_INCLUDE,
-  STATE_PACKAGE,
-  STATE_NAMESPACE, /* 5 */
+  STATE_C_INCLUDE,
+  STATE_PACKAGE,  /* 5 */
+  STATE_NAMESPACE,
   STATE_ENUM,
   STATE_BITFIELD,
   STATE_FUNCTION,
-  STATE_FUNCTION_RETURN,
-  STATE_FUNCTION_PARAMETERS, /* 10 */
+  STATE_FUNCTION_RETURN,  /* 10 */
+  STATE_FUNCTION_PARAMETERS,
   STATE_FUNCTION_PARAMETER,
   STATE_CLASS,
   STATE_CLASS_FIELD,
-  STATE_CLASS_PROPERTY,
-  STATE_INTERFACE, /* 15 */
+  STATE_CLASS_PROPERTY,  /* 15 */
+  STATE_INTERFACE,
   STATE_INTERFACE_PROPERTY,
   STATE_INTERFACE_FIELD,
   STATE_IMPLEMENTS,
-  STATE_PREREQUISITE,
-  STATE_BOXED,   /* 20 */
+  STATE_PREREQUISITE,    /* 20 */
+  STATE_BOXED,
   STATE_BOXED_FIELD,
   STATE_STRUCT,
   STATE_STRUCT_FIELD,
-  STATE_ERRORDOMAIN,
-  STATE_UNION, /* 25 */
+  STATE_ERRORDOMAIN, /* 25 */
+  STATE_UNION,
   STATE_UNION_FIELD,
   STATE_NAMESPACE_CONSTANT,
   STATE_CLASS_CONSTANT,
-  STATE_INTERFACE_CONSTANT,
-  STATE_ALIAS, /* 30 */
+  STATE_INTERFACE_CONSTANT,  /* 30 */
+  STATE_ALIAS,
   STATE_TYPE,
   STATE_ATTRIBUTE,
   STATE_DOC,
@@ -100,6 +101,7 @@ struct _ParseContext
   const char *c_prefix;
   GIrModule *current_module;
   GSList *node_stack;
+  char *current_alias;
   GIrNode *current_typed;
   GList *type_stack;
   GList *type_parameters;
@@ -126,6 +128,8 @@ static void text_handler          (GMarkupParseContext *context,
 static void cleanup               (GMarkupParseContext *context,
 				   GError              *error,
 				   gpointer             user_data);
+static void state_switch (ParseContext *ctx, ParseState newstate);
+
 
 static GMarkupParser markup_parser =
 {
@@ -143,6 +147,13 @@ start_alias (GMarkupParseContext *context,
 	     const gchar        **attribute_values,
 	     ParseContext        *ctx,
 	     GError             **error);
+static gboolean
+start_type (GMarkupParseContext *context,
+	    const gchar         *element_name,
+	    const gchar        **attribute_names,
+	    const gchar        **attribute_values,
+	    ParseContext        *ctx,
+	    GError             **error);
 
 static const gchar *find_attribute (const gchar  *name,
 				    const gchar **attribute_names,
@@ -150,7 +161,7 @@ static const gchar *find_attribute (const gchar  *name,
 
 
 GIrParser *
-g_ir_parser_new (void)
+_g_ir_parser_new (void)
 {
   GIrParser *parser = g_slice_new0 (GIrParser);
 
@@ -158,7 +169,7 @@ g_ir_parser_new (void)
 }
 
 void
-g_ir_parser_free (GIrParser *parser)
+_g_ir_parser_free (GIrParser *parser)
 {
   GList *l;
 
@@ -166,14 +177,14 @@ g_ir_parser_free (GIrParser *parser)
     g_strfreev (parser->includes);
 
   for (l = parser->parsed_modules; l; l = l->next)
-    g_ir_module_free (l->data);
+    _g_ir_module_free (l->data);
 
   g_slice_free (GIrParser, parser);
 }
 
 void
-g_ir_parser_set_includes (GIrParser          *parser,
-			  const gchar *const *includes)
+_g_ir_parser_set_includes (GIrParser          *parser,
+			   const gchar *const *includes)
 {
   if (parser->includes)
     g_strfreev (parser->includes);
@@ -196,6 +207,11 @@ firstpass_start_element_handler (GMarkupParseContext *context,
       start_alias (context, element_name, attribute_names, attribute_values,
 		   ctx, error);
     }
+  else if (ctx->state == STATE_ALIAS && strcmp (element_name, "type") == 0)
+    {
+      start_type (context, element_name, attribute_names, attribute_values,
+		  ctx, error);
+    }
   else if (strcmp (element_name, "record") == 0)
     {
       const gchar *name;
@@ -220,6 +236,15 @@ firstpass_end_element_handler (GMarkupParseContext *context,
 			       gpointer             user_data,
 			       GError             **error)
 {
+  ParseContext *ctx = user_data;
+  if (strcmp (element_name, "alias") == 0)
+    {
+      state_switch (ctx, STATE_NAMESPACE);
+      g_free (ctx->current_alias);
+      ctx->current_alias = NULL;
+    }
+  else if (strcmp (element_name, "type") == 0 && ctx->state == STATE_TYPE)
+    state_switch (ctx, ctx->prev_state);
 }
 
 static GMarkupParser firstpass_parser =
@@ -268,7 +293,7 @@ locate_gir (GIrParser  *parser,
   return NULL;
 }
 
-#define MISSING_ATTRIBUTE(ctx,error,element,attribute)			        \
+#define MISSING_ATTRIBUTE(context,error,element,attribute)			        \
   do {                                                                          \
     int line_number, char_number;                                                \
     g_markup_parse_context_get_position (context, &line_number, &char_number);  \
@@ -323,7 +348,8 @@ push_node (ParseContext *ctx, GIrNode *node)
   ctx->node_stack = g_slist_prepend (ctx->node_stack, node);
 }
 
-static GIrNodeType * parse_type_internal (const gchar *str, gchar **next, gboolean in_glib,
+static GIrNodeType * parse_type_internal (GIrModule *module,
+					  const gchar *str, gchar **next, gboolean in_glib,
 					  gboolean in_gobject);
 
 typedef struct {
@@ -333,15 +359,17 @@ typedef struct {
 } IntegerAliasInfo;
 
 static IntegerAliasInfo integer_aliases[] = {
-  { "char",     SIZEOF_CHAR,      0 },
-  { "short",    SIZEOF_SHORT,     1 },
-  { "ushort",   SIZEOF_SHORT,     0 },
-  { "int",      SIZEOF_INT,       1 },
-  { "uint",     SIZEOF_INT,       0 },
-  { "long",     SIZEOF_LONG,      1 },
-  { "ulong",    SIZEOF_LONG,      0 },
-  { "gsize",    GLIB_SIZEOF_SIZE_T,    0 },
+  { "gchar",    SIZEOF_CHAR,      0 },
+  { "gshort",   SIZEOF_SHORT,     1 },
+  { "gushort",  SIZEOF_SHORT,     0 },
+  { "gint",     SIZEOF_INT,       1 },
+  { "guint",    SIZEOF_INT,       0 },
+  { "glong",    SIZEOF_LONG,      1 },
+  { "gulong",   SIZEOF_LONG,      0 },
   { "gssize",   GLIB_SIZEOF_SIZE_T,    1 },
+  { "gsize",    GLIB_SIZEOF_SIZE_T,    0 },
+  { "gintptr",  GLIB_SIZEOF_SIZE_T,    1 },
+  { "guintptr", GLIB_SIZEOF_SIZE_T,    0 },
 };
 
 typedef struct {
@@ -353,39 +381,40 @@ typedef struct {
 #define BASIC_TYPE_FIXED_OFFSET 3
 
 static BasicTypeInfo basic_types[] = {
-    { "none",     GI_TYPE_TAG_VOID,    0 },
-    { "any",      GI_TYPE_TAG_VOID,    1 },
+    { "none",      GI_TYPE_TAG_VOID,    0 },
+    { "gpointer",  GI_TYPE_TAG_VOID,    1 },
 
-    { "bool",     GI_TYPE_TAG_BOOLEAN, 0 },
-    { "int8",     GI_TYPE_TAG_INT8,    0 }, /* Start of BASIC_TYPE_FIXED_OFFSET */
-    { "uint8",    GI_TYPE_TAG_UINT8,   0 },
-    { "int16",    GI_TYPE_TAG_INT16,   0 },
-    { "uint16",   GI_TYPE_TAG_UINT16,  0 },
-    { "int32",    GI_TYPE_TAG_INT32,   0 },
-    { "uint32",   GI_TYPE_TAG_UINT32,  0 },
-    { "int64",    GI_TYPE_TAG_INT64,   0 },
-    { "uint64",   GI_TYPE_TAG_UINT64,  0 },
-    { "float",    GI_TYPE_TAG_FLOAT,   0 },
-    { "double",   GI_TYPE_TAG_DOUBLE,  0 },
-    { "GType",    GI_TYPE_TAG_GTYPE,   0 },
-    { "utf8",     GI_TYPE_TAG_UTF8,    1 },
-    { "filename", GI_TYPE_TAG_FILENAME,1 },
+    { "gboolean",  GI_TYPE_TAG_BOOLEAN, 0 },
+    { "gint8",     GI_TYPE_TAG_INT8,    0 }, /* Start of BASIC_TYPE_FIXED_OFFSET */
+    { "guint8",    GI_TYPE_TAG_UINT8,   0 },
+    { "gint16",    GI_TYPE_TAG_INT16,   0 },
+    { "guint16",   GI_TYPE_TAG_UINT16,  0 },
+    { "gint32",    GI_TYPE_TAG_INT32,   0 },
+    { "guint32",   GI_TYPE_TAG_UINT32,  0 },
+    { "gint64",    GI_TYPE_TAG_INT64,   0 },
+    { "guint64",   GI_TYPE_TAG_UINT64,  0 },
+    { "gfloat",    GI_TYPE_TAG_FLOAT,   0 },
+    { "gdouble",   GI_TYPE_TAG_DOUBLE,  0 },
+    { "GType",     GI_TYPE_TAG_GTYPE,   0 },
+    { "utf8",      GI_TYPE_TAG_UTF8,    1 },
+    { "filename",  GI_TYPE_TAG_FILENAME,1 },
+    { "gunichar",  GI_TYPE_TAG_UNICHAR, 0 },
 };
 
 static const BasicTypeInfo *
 parse_basic (const char *str)
 {
-  gint i;
-  gint n_basic = G_N_ELEMENTS (basic_types);
+  guint i;
+  guint n_basic = G_N_ELEMENTS (basic_types);
 
   for (i = 0; i < n_basic; i++)
     {
-      if (g_str_has_prefix (str, basic_types[i].str))
+      if (strcmp (str, basic_types[i].str) == 0)
 	return &(basic_types[i]);
     }
   for (i = 0; i < G_N_ELEMENTS (integer_aliases); i++)
     {
-      if (g_str_has_prefix (str, integer_aliases[i].str))
+      if (strcmp (str, integer_aliases[i].str) == 0)
 	{
 	  switch (integer_aliases[i].size)
 	    {
@@ -422,14 +451,15 @@ parse_basic (const char *str)
 }
 
 static GIrNodeType *
-parse_type_internal (const gchar *str, char **next, gboolean in_glib,
+parse_type_internal (GIrModule *module,
+		     const gchar *str, char **next, gboolean in_glib,
 		     gboolean in_gobject)
 {
   const BasicTypeInfo *basic;
   GIrNodeType *type;
   char *temporary_type = NULL;
 
-  type = (GIrNodeType *)g_ir_node_new (G_IR_NODE_TYPE);
+  type = (GIrNodeType *)_g_ir_node_new (G_IR_NODE_TYPE, module);
 
   type->unparsed = g_strdup (str);
 
@@ -552,12 +582,12 @@ parse_type_internal (const gchar *str, char **next, gboolean in_glib,
 
   if (next)
     *next = (char*)str;
-  g_assert (type->tag >= 0 && type->tag <= GI_TYPE_TAG_ERROR);
+  g_assert (type->tag >= 0 && type->tag < GI_TYPE_TAG_N_TYPES);
   g_free (temporary_type);
   return type;
 
 /* error: */
-  g_ir_node_free ((GIrNode *)type);
+  _g_ir_node_free ((GIrNode *)type);
   g_free (temporary_type);
   return NULL;
 }
@@ -643,7 +673,7 @@ parse_type (ParseContext *ctx, const gchar *type)
   if (basic == NULL)
     type = resolve_aliases (ctx, type);
 
-  node = parse_type_internal (type, NULL, in_glib, in_gobject);
+  node = parse_type_internal (ctx->current_module, type, NULL, in_glib, in_gobject);
   if (node)
     g_debug ("Parsed type: %s => %d", type, node->tag);
   else
@@ -660,12 +690,15 @@ introspectable_prelude (GMarkupParseContext *context,
 		    ParseState           new_state)
 {
   const gchar *introspectable_arg;
+  const gchar *shadowed_by;
   gboolean introspectable;
 
   g_assert (ctx->state != STATE_PASSTHROUGH);
 
   introspectable_arg = find_attribute ("introspectable", attribute_names, attribute_values);
-  introspectable = !(introspectable_arg && atoi (introspectable_arg) == 0);
+  shadowed_by = find_attribute ("shadowed-by", attribute_names, attribute_values);
+
+  introspectable = !(introspectable_arg && atoi (introspectable_arg) == 0) && shadowed_by == NULL;
 
   if (introspectable)
     state_switch (ctx, new_state);
@@ -720,7 +753,8 @@ start_glib_boxed (GMarkupParseContext *context,
       return FALSE;
     }
 
-  boxed = (GIrNodeBoxed *) g_ir_node_new (G_IR_NODE_BOXED);
+  boxed = (GIrNodeBoxed *) _g_ir_node_new (G_IR_NODE_BOXED,
+					  ctx->current_module);
 
   ((GIrNode *)boxed)->name = g_strdup (name);
   boxed->gtype_name = g_strdup (typename);
@@ -746,11 +780,13 @@ start_function (GMarkupParseContext *context,
 		GError             **error)
 {
   const gchar *name;
+  const gchar *shadows;
   const gchar *symbol;
   const gchar *deprecated;
   const gchar *throws;
   GIrNodeFunction *function;
   gboolean found = FALSE;
+  gboolean in_embedded_type;
 
   switch (ctx->state)
     {
@@ -759,15 +795,14 @@ start_function (GMarkupParseContext *context,
 	       strcmp (element_name, "callback") == 0);
       break;
     case STATE_CLASS:
-      found = strcmp (element_name, "function") == 0;
-	/* fallthrough */
     case STATE_BOXED:
     case STATE_STRUCT:
     case STATE_UNION:
-      found = (found || strcmp (element_name, "constructor") == 0);
+      found = strcmp (element_name, "constructor") == 0;
       /* fallthrough */
     case STATE_INTERFACE:
       found = (found ||
+	       strcmp (element_name, "function") == 0 ||
 	       strcmp (element_name, "method") == 0 ||
 	       strcmp (element_name, "callback") == 0);
       break;
@@ -781,13 +816,15 @@ start_function (GMarkupParseContext *context,
   if (!found)
     return FALSE;
 
-  if (ctx->state == STATE_STRUCT_FIELD)
-    ctx->in_embedded_type = TRUE;
+  in_embedded_type = ctx->state == STATE_STRUCT_FIELD;
 
   if (!introspectable_prelude (context, attribute_names, attribute_values, ctx, STATE_FUNCTION))
     return TRUE;
 
+  ctx->in_embedded_type = in_embedded_type;
+
   name = find_attribute ("name", attribute_names, attribute_values);
+  shadows = find_attribute ("shadows", attribute_names, attribute_values);
   symbol = find_attribute ("c:identifier", attribute_names, attribute_values);
   deprecated = find_attribute ("deprecated", attribute_names, attribute_values);
   throws = find_attribute ("throws", attribute_names, attribute_values);
@@ -803,7 +840,11 @@ start_function (GMarkupParseContext *context,
       return FALSE;
     }
 
-  function = (GIrNodeFunction *) g_ir_node_new (G_IR_NODE_FUNCTION);
+  if (shadows)
+    name = shadows;
+
+  function = (GIrNodeFunction *) _g_ir_node_new (G_IR_NODE_FUNCTION,
+						ctx->current_module);
 
   ((GIrNode *)function)->name = g_strdup (name);
   function->symbol = g_strdup (symbol);
@@ -935,15 +976,16 @@ parse_property_transfer (GIrNodeProperty *property,
     }
 }
 
-static void
-parse_param_transfer (GIrNodeParam *param, const gchar *transfer, const gchar *name)
+static gboolean
+parse_param_transfer (GIrNodeParam *param, const gchar *transfer, const gchar *name,
+		      GError **error)
 {
   if (transfer == NULL)
   {
-    if (!name)
-      g_warning ("required attribute 'transfer-ownership' missing");
-    else
-      g_warning ("required attribute 'transfer-ownership' for function '%s'", name);
+    g_set_error (error, G_MARKUP_ERROR,
+		 G_MARKUP_ERROR_INVALID_CONTENT,
+		 "required attribute 'transfer-ownership' missing");
+    return FALSE;
   }
   else if (strcmp (transfer, "none") == 0)
     {
@@ -962,8 +1004,12 @@ parse_param_transfer (GIrNodeParam *param, const gchar *transfer, const gchar *n
     }
   else
     {
-      g_warning ("Unknown transfer-ownership value: %s", transfer);
+      g_set_error (error, G_MARKUP_ERROR,
+		   G_MARKUP_ERROR_INVALID_CONTENT,
+		   "invalid value for 'transfer-ownership': %s", transfer);
+      return FALSE;
     }
+  return TRUE;
 }
 
 static gboolean
@@ -1004,7 +1050,8 @@ start_parameter (GMarkupParseContext *context,
   if (name == NULL)
     name = "unknown";
 
-  param = (GIrNodeParam *)g_ir_node_new (G_IR_NODE_PARAM);
+  param = (GIrNodeParam *)_g_ir_node_new (G_IR_NODE_PARAM,
+					  ctx->current_module);
 
   ctx->current_typed = (GIrNode*) param;
   ctx->current_typed->name = g_strdup (name);
@@ -1048,7 +1095,8 @@ start_parameter (GMarkupParseContext *context,
   else
     param->allow_none = FALSE;
 
-  parse_param_transfer (param, transfer, name);
+  if (!parse_param_transfer (param, transfer, name, error))
+    return FALSE;
 
   if (scope && strcmp (scope, "call") == 0)
     param->scope = GI_SCOPE_TYPE_CALL;
@@ -1112,14 +1160,25 @@ start_field (GMarkupParseContext *context,
   const gchar *bits;
   const gchar *branch;
   GIrNodeField *field;
+  ParseState target_state;
+  gboolean introspectable;
 
   switch (ctx->state)
     {
     case STATE_CLASS:
+      target_state = STATE_CLASS_FIELD;
+      break;
     case STATE_BOXED:
+      target_state = STATE_BOXED_FIELD;
+      break;
     case STATE_STRUCT:
+      target_state = STATE_STRUCT_FIELD;
+      break;
     case STATE_UNION:
+      target_state = STATE_UNION_FIELD;
+      break;
     case STATE_INTERFACE:
+      target_state = STATE_INTERFACE_FIELD;
       break;
     default:
       return FALSE;
@@ -1128,6 +1187,13 @@ start_field (GMarkupParseContext *context,
   if (strcmp (element_name, "field") != 0)
     return FALSE;
 
+  g_assert (ctx->state != STATE_PASSTHROUGH);
+
+  /* We handle introspectability specially here; we replace with just gpointer
+   * for the type.
+   */
+  introspectable = introspectable_prelude (context, attribute_names, attribute_values, ctx, target_state);
+
   name = find_attribute ("name", attribute_names, attribute_values);
   readable = find_attribute ("readable", attribute_names, attribute_values);
   writable = find_attribute ("writable", attribute_names, attribute_values);
@@ -1140,8 +1206,17 @@ start_field (GMarkupParseContext *context,
       return FALSE;
     }
 
-  field = (GIrNodeField *)g_ir_node_new (G_IR_NODE_FIELD);
-  ctx->current_typed = (GIrNode*) field;
+  field = (GIrNodeField *)_g_ir_node_new (G_IR_NODE_FIELD,
+					  ctx->current_module);
+  if (introspectable)
+    {
+      ctx->current_typed = (GIrNode*) field;
+    }
+  else
+    {
+      field->type = parse_type (ctx, "gpointer");
+    }
+
   ((GIrNode *)field)->name = g_strdup (name);
   /* Fields are assumed to be read-only.
    * (see also girwriter.py and generate.c)
@@ -1162,7 +1237,6 @@ start_field (GMarkupParseContext *context,
 
 	iface = (GIrNodeInterface *)CURRENT_NODE (ctx);
 	iface->members = g_list_append (iface->members, field);
-	state_switch (ctx, STATE_CLASS_FIELD);
       }
       break;
     case G_IR_NODE_INTERFACE:
@@ -1171,7 +1245,6 @@ start_field (GMarkupParseContext *context,
 
 	iface = (GIrNodeInterface *)CURRENT_NODE (ctx);
 	iface->members = g_list_append (iface->members, field);
-	state_switch (ctx, STATE_INTERFACE_FIELD);
       }
       break;
     case G_IR_NODE_BOXED:
@@ -1180,7 +1253,6 @@ start_field (GMarkupParseContext *context,
 
 	boxed = (GIrNodeBoxed *)CURRENT_NODE (ctx);
 		boxed->members = g_list_append (boxed->members, field);
-		state_switch (ctx, STATE_BOXED_FIELD);
       }
       break;
     case G_IR_NODE_STRUCT:
@@ -1189,7 +1261,6 @@ start_field (GMarkupParseContext *context,
 
 	struct_ = (GIrNodeStruct *)CURRENT_NODE (ctx);
 	struct_->members = g_list_append (struct_->members, field);
-	state_switch (ctx, STATE_STRUCT_FIELD);
       }
       break;
     case G_IR_NODE_UNION:
@@ -1202,7 +1273,8 @@ start_field (GMarkupParseContext *context,
 	  {
 	    GIrNodeConstant *constant;
 
-	    constant = (GIrNodeConstant *) g_ir_node_new (G_IR_NODE_CONSTANT);
+	    constant = (GIrNodeConstant *) _g_ir_node_new (G_IR_NODE_CONSTANT,
+							  ctx->current_module);
 	    ((GIrNode *)constant)->name = g_strdup (name);
 	    constant->value = g_strdup (branch);
 	    constant->type = union_->discriminator_type;
@@ -1210,7 +1282,6 @@ start_field (GMarkupParseContext *context,
 
 	    union_->discriminators = g_list_append (union_->discriminators, constant);
 	  }
-	state_switch (ctx, STATE_UNION_FIELD);
       }
       break;
     default:
@@ -1229,9 +1300,6 @@ start_alias (GMarkupParseContext *context,
 	     GError             **error)
 {
   const gchar *name;
-  const gchar *target;
-  char *key;
-  char *value;
 
   name = find_attribute ("name", attribute_names, attribute_values);
   if (name == NULL)
@@ -1240,26 +1308,8 @@ start_alias (GMarkupParseContext *context,
       return FALSE;
     }
 
-  target = find_attribute ("target", attribute_names, attribute_values);
-  if (name == NULL)
-    {
-      MISSING_ATTRIBUTE (context, error, element_name, "target");
-      return FALSE;
-    }
-
-  value = g_strdup (target);
-  key = g_strdup_printf ("%s.%s", ctx->namespace, name);
-  if (!strchr (target, '.'))
-    {
-      const BasicTypeInfo *basic = parse_basic (target);
-      if (!basic)
-	{
-	  g_free (value);
-	  /* For non-basic types, re-qualify the interface */
-	  value = g_strdup_printf ("%s.%s", ctx->namespace, target);
-	}
-    }
-  g_hash_table_replace (ctx->aliases, key, value);
+  ctx->current_alias = g_strdup (name);
+  state_switch (ctx, STATE_ALIAS);
 
   return TRUE;
 }
@@ -1297,9 +1347,11 @@ start_enum (GMarkupParseContext *context,
     }
 
   if (strcmp (element_name, "enumeration") == 0)
-    enum_ = (GIrNodeEnum *) g_ir_node_new (G_IR_NODE_ENUM);
+    enum_ = (GIrNodeEnum *) _g_ir_node_new (G_IR_NODE_ENUM,
+					   ctx->current_module);
   else
-    enum_ = (GIrNodeEnum *) g_ir_node_new (G_IR_NODE_FLAGS);
+    enum_ = (GIrNodeEnum *) _g_ir_node_new (G_IR_NODE_FLAGS,
+					   ctx->current_module);
   ((GIrNode *)enum_)->name = g_strdup (name);
   enum_->gtype_name = g_strdup (typename);
   enum_->gtype_init = g_strdup (typeinit);
@@ -1362,7 +1414,8 @@ start_property (GMarkupParseContext *context,
       return FALSE;
     }
 
-  property = (GIrNodeProperty *) g_ir_node_new (G_IR_NODE_PROPERTY);
+  property = (GIrNodeProperty *) _g_ir_node_new (G_IR_NODE_PROPERTY,
+						ctx->current_module);
   ctx->current_typed = (GIrNode*) property;
 
   ((GIrNode *)property)->name = g_strdup (name);
@@ -1393,7 +1446,7 @@ start_property (GMarkupParseContext *context,
   return TRUE;
 }
 
-static gint
+static gint64
 parse_value (const gchar *str)
 {
   gchar *shift_op;
@@ -1403,15 +1456,15 @@ parse_value (const gchar *str)
 
   if (shift_op)
     {
-      gint base, shift;
+      gint64 base, shift;
 
-      base = strtol (str, NULL, 10);
-      shift = strtol (shift_op + 3, NULL, 10);
+      base = g_ascii_strtoll (str, NULL, 10);
+      shift = g_ascii_strtoll (shift_op + 3, NULL, 10);
 
       return base << shift;
     }
   else
-    return strtol (str, NULL, 10);
+    return g_ascii_strtoll (str, NULL, 10);
 
   return 0;
 }
@@ -1427,6 +1480,7 @@ start_member (GMarkupParseContext *context,
   const gchar *name;
   const gchar *value;
   const gchar *deprecated;
+  const gchar *c_identifier;
   GIrNodeEnum *enum_;
   GIrNodeValue *value_;
 
@@ -1437,6 +1491,7 @@ start_member (GMarkupParseContext *context,
   name = find_attribute ("name", attribute_names, attribute_values);
   value = find_attribute ("value", attribute_names, attribute_values);
   deprecated = find_attribute ("deprecated", attribute_names, attribute_values);
+  c_identifier = find_attribute ("c:identifier", attribute_names, attribute_values);
 
   if (name == NULL)
     {
@@ -1444,7 +1499,8 @@ start_member (GMarkupParseContext *context,
       return FALSE;
     }
 
-  value_ = (GIrNodeValue *) g_ir_node_new (G_IR_NODE_VALUE);
+  value_ = (GIrNodeValue *) _g_ir_node_new (G_IR_NODE_VALUE,
+					   ctx->current_module);
 
   ((GIrNode *)value_)->name = g_strdup (name);
 
@@ -1455,6 +1511,10 @@ start_member (GMarkupParseContext *context,
   else
     value_->deprecated = FALSE;
 
+  g_hash_table_insert (((GIrNode *)value_)->attributes,
+                       g_strdup ("c:identifier"),
+                       g_strdup (c_identifier));
+
   enum_ = (GIrNodeEnum *)CURRENT_NODE (ctx);
   enum_->values = g_list_append (enum_->values, value_);
 
@@ -1517,7 +1577,8 @@ start_constant (GMarkupParseContext *context,
       return FALSE;
     }
 
-  constant = (GIrNodeConstant *) g_ir_node_new (G_IR_NODE_CONSTANT);
+  constant = (GIrNodeConstant *) _g_ir_node_new (G_IR_NODE_CONSTANT,
+						ctx->current_module);
 
   ((GIrNode *)constant)->name = g_strdup (name);
   constant->value = g_strdup (value);
@@ -1589,7 +1650,8 @@ start_errordomain (GMarkupParseContext *context,
       return FALSE;
     }
 
-  domain = (GIrNodeErrorDomain *) g_ir_node_new (G_IR_NODE_ERROR_DOMAIN);
+  domain = (GIrNodeErrorDomain *) _g_ir_node_new (G_IR_NODE_ERROR_DOMAIN,
+						 ctx->current_module);
 
   ((GIrNode *)domain)->name = g_strdup (name);
   domain->getquark = g_strdup (getquark);
@@ -1651,7 +1713,8 @@ start_interface (GMarkupParseContext *context,
       return FALSE;
     }
 
-  iface = (GIrNodeInterface *) g_ir_node_new (G_IR_NODE_INTERFACE);
+  iface = (GIrNodeInterface *) _g_ir_node_new (G_IR_NODE_INTERFACE,
+					      ctx->current_module);
   ((GIrNode *)iface)->name = g_strdup (name);
   iface->gtype_name = g_strdup (typename);
   iface->gtype_init = g_strdup (typeinit);
@@ -1726,7 +1789,8 @@ start_class (GMarkupParseContext *context,
       return FALSE;
     }
 
-  iface = (GIrNodeInterface *) g_ir_node_new (G_IR_NODE_OBJECT);
+  iface = (GIrNodeInterface *) _g_ir_node_new (G_IR_NODE_OBJECT,
+					      ctx->current_module);
   ((GIrNode *)iface)->name = g_strdup (name);
   iface->gtype_name = g_strdup (typename);
   iface->gtype_init = g_strdup (typeinit);
@@ -1739,6 +1803,8 @@ start_class (GMarkupParseContext *context,
 
   iface->abstract = abstract && strcmp (abstract, "1") == 0;
 
+  if (fundamental)
+    iface->fundamental = TRUE;
   if (ref_func)
     iface->ref_func = g_strdup (ref_func);
   if (unref_func)
@@ -1765,6 +1831,7 @@ start_type (GMarkupParseContext *context,
 {
   const gchar *name;
   const gchar *ctype;
+  gboolean in_alias = FALSE;
   gboolean is_array;
   gboolean is_varargs;
   GIrNodeType *typenode;
@@ -1792,15 +1859,55 @@ start_type (GMarkupParseContext *context,
 	   ctx->state == STATE_BOXED_FIELD ||
 	   ctx->state == STATE_NAMESPACE_CONSTANT ||
 	   ctx->state == STATE_CLASS_CONSTANT ||
-	   ctx->state == STATE_INTERFACE_CONSTANT
+	   ctx->state == STATE_INTERFACE_CONSTANT ||
+	   ctx->state == STATE_ALIAS
 	   )
     {
+      if (ctx->state == STATE_ALIAS)
+	in_alias = TRUE;
       state_switch (ctx, STATE_TYPE);
       ctx->type_depth = 1;
       ctx->type_stack = NULL;
       ctx->type_parameters = NULL;
     }
 
+  name = find_attribute ("name", attribute_names, attribute_values);
+
+  if (in_alias && ctx->current_alias)
+    {
+      char *key;
+      char *value;
+
+      if (name == NULL)
+	{
+	  MISSING_ATTRIBUTE (context, error, element_name, "name");
+	  return FALSE;
+	}
+
+      key = g_strdup_printf ("%s.%s", ctx->namespace, ctx->current_alias);
+      if (!strchr (name, '.'))
+	{
+	  const BasicTypeInfo *basic = parse_basic (name);
+	  if (!basic)
+	    {
+	      /* For non-basic types, re-qualify the interface */
+	      value = g_strdup_printf ("%s.%s", ctx->namespace, name);
+	    }
+	  else
+	    {
+	      value = g_strdup (name);
+	    }
+	}
+      else
+	value = g_strdup (name);
+
+      g_hash_table_replace (ctx->aliases, key, value);
+
+      return TRUE;
+    }
+  else if (!ctx->current_module || in_alias)
+    return TRUE;
+
   if (!ctx->current_typed)
     {
       g_set_error (error,
@@ -1819,13 +1926,13 @@ start_type (GMarkupParseContext *context,
       const char *len;
       const char *size;
 
-      typenode = (GIrNodeType *)g_ir_node_new (G_IR_NODE_TYPE);
+      typenode = (GIrNodeType *)_g_ir_node_new (G_IR_NODE_TYPE,
+						ctx->current_module);
 
       typenode->tag = GI_TYPE_TAG_ARRAY;
       typenode->is_pointer = TRUE;
       typenode->is_array = TRUE;
 
-      name = find_attribute ("name", attribute_names, attribute_values);
       if (name && strcmp (name, "GLib.Array") == 0) {
         typenode->array_type = GI_ARRAY_TYPE_ARRAY;
       } else if (name && strcmp (name, "GLib.ByteArray") == 0) {
@@ -1866,10 +1973,12 @@ start_type (GMarkupParseContext *context,
   else
     {
       int pointer_depth;
-      name = find_attribute ("name", attribute_names, attribute_values);
 
       if (name == NULL)
-	MISSING_ATTRIBUTE (context, error, element_name, "name");
+	{
+	  MISSING_ATTRIBUTE (context, error, element_name, "name");
+	  return FALSE;
+	}
 
       pointer_depth = 0;
       ctype = find_attribute ("c:type", attribute_names, attribute_values);
@@ -1919,14 +2028,14 @@ end_type_top (ParseContext *ctx)
       typenode->tag == GI_TYPE_TAG_GSLIST)
     {
       if (typenode->parameter_type1 == NULL)
-	typenode->parameter_type1 = parse_type (ctx, "any");
+	typenode->parameter_type1 = parse_type (ctx, "gpointer");
     }
   else if (typenode->tag == GI_TYPE_TAG_GHASH)
     {
       if (typenode->parameter_type1 == NULL)
 	{
-	  typenode->parameter_type1 = parse_type (ctx, "any");
-	  typenode->parameter_type2 = parse_type (ctx, "any");
+	  typenode->parameter_type1 = parse_type (ctx, "gpointer");
+	  typenode->parameter_type2 = parse_type (ctx, "gpointer");
 	}
     }
 
@@ -2028,7 +2137,7 @@ start_doc (GMarkupParseContext *context,
 	   ParseContext       *ctx,
 	   GError             **error)
 {
-  if (strcmp (element_name, "doc") != 0 || ctx->node_stack == NULL)
+  if (strcmp (element_name, "doc") != 0)
     return FALSE;
 
   state_switch (ctx, STATE_DOC);
@@ -2096,7 +2205,8 @@ start_return_value (GMarkupParseContext *context,
 	ctx->state == STATE_FUNCTION))
     return FALSE;
 
-  param = (GIrNodeParam *)g_ir_node_new (G_IR_NODE_PARAM);
+  param = (GIrNodeParam *)_g_ir_node_new (G_IR_NODE_PARAM,
+					  ctx->current_module);
   param->in = FALSE;
   param->out = FALSE;
   param->retval = TRUE;
@@ -2106,7 +2216,8 @@ start_return_value (GMarkupParseContext *context,
   state_switch (ctx, STATE_FUNCTION_RETURN);
 
   transfer = find_attribute ("transfer-ownership", attribute_names, attribute_values);
-  parse_param_transfer (param, transfer, NULL);
+  if (!parse_param_transfer (param, transfer, NULL, error))
+    return FALSE;
 
   switch (CURRENT_NODE (ctx)->type)
     {
@@ -2205,7 +2316,8 @@ start_glib_signal (GMarkupParseContext *context,
       MISSING_ATTRIBUTE (context, error, element_name, "name");
       return FALSE;
     }
-  signal = (GIrNodeSignal *)g_ir_node_new (G_IR_NODE_SIGNAL);
+  signal = (GIrNodeSignal *)_g_ir_node_new (G_IR_NODE_SIGNAL,
+					    ctx->current_module);
 
   ((GIrNode *)signal)->name = g_strdup (name);
 
@@ -2286,7 +2398,8 @@ start_vfunc (GMarkupParseContext *context,
       return FALSE;
     }
 
-  vfunc = (GIrNodeVFunc *)g_ir_node_new (G_IR_NODE_VFUNC);
+  vfunc = (GIrNodeVFunc *)_g_ir_node_new (G_IR_NODE_VFUNC,
+					  ctx->current_module);
 
   ((GIrNode *)vfunc)->name = g_strdup (name);
 
@@ -2319,7 +2432,7 @@ start_vfunc (GMarkupParseContext *context,
   if (offset)
     vfunc->offset = atoi (offset);
   else
-    vfunc->offset = 0;
+    vfunc->offset = 0xFFFF;
 
   vfunc->invoker = g_strdup (invoker);
 
@@ -2382,7 +2495,8 @@ start_struct (GMarkupParseContext *context,
       return FALSE;
     }
 
-  struct_ = (GIrNodeStruct *) g_ir_node_new (G_IR_NODE_STRUCT);
+  struct_ = (GIrNodeStruct *) _g_ir_node_new (G_IR_NODE_STRUCT,
+					     ctx->current_module);
 
   ((GIrNode *)struct_)->name = g_strdup (name ? name : "");
   if (deprecated)
@@ -2442,7 +2556,8 @@ start_union (GMarkupParseContext *context,
       return FALSE;
     }
 
-  union_ = (GIrNodeUnion *) g_ir_node_new (G_IR_NODE_UNION);
+  union_ = (GIrNodeUnion *) _g_ir_node_new (G_IR_NODE_UNION,
+					   ctx->current_module);
 
   ((GIrNode *)union_)->name = g_strdup (name ? name : "");
   union_->gtype_name = g_strdup (typename);
@@ -2504,7 +2619,7 @@ parse_include (GMarkupParseContext *context,
   gchar *buffer;
   gsize length;
   gchar *girpath, *girname;
-  GList *modules;
+  GIrModule *module;
   GList *l;
 
   for (l = ctx->parser->parsed_modules; l; l = l->next)
@@ -2540,7 +2655,7 @@ parse_include (GMarkupParseContext *context,
     }
   g_free (girname);
 
-  g_debug ("Parsing include %s", girpath);
+  g_debug ("Parsing include %s\n", girpath);
 
   if (!g_file_get_contents (girpath, &buffer, &length, &error))
     {
@@ -2550,7 +2665,7 @@ parse_include (GMarkupParseContext *context,
       return FALSE;
     }
 
-  modules = g_ir_parser_parse_string (ctx->parser, name, girpath, buffer, length, &error);
+  module = _g_ir_parser_parse_string (ctx->parser, name, girpath, buffer, length, &error);
   g_free (buffer);
   if (error != NULL)
     {
@@ -2563,8 +2678,8 @@ parse_include (GMarkupParseContext *context,
     }
   g_free (girpath);
 
-  ctx->include_modules = g_list_concat (ctx->include_modules,
-					modules);
+  ctx->include_modules = g_list_append (ctx->include_modules,
+					module);
 
   return TRUE;
 }
@@ -2734,6 +2849,11 @@ start_element_handler (GMarkupParseContext *context,
 				 attribute_names, attribute_values,
 				 ctx, error))
 	goto out;
+      else if (strcmp (element_name, "c:include") == 0)
+	{
+	  state_switch (ctx, STATE_C_INCLUDE);
+	  goto out;
+	}
       break;
 
     case 'm':
@@ -2764,7 +2884,7 @@ start_element_handler (GMarkupParseContext *context,
 	  name = find_attribute ("name", attribute_names, attribute_values);
 	  version = find_attribute ("version", attribute_names, attribute_values);
 	  shared_library = find_attribute ("shared-library", attribute_names, attribute_values);
-	  cprefix = find_attribute ("c:prefix", attribute_names, attribute_values);
+	  cprefix = find_attribute ("c:identifier-prefixes", attribute_names, attribute_values);
 
 	  if (name == NULL)
 	    MISSING_ATTRIBUTE (context, error, element_name, "name");
@@ -2781,7 +2901,7 @@ start_element_handler (GMarkupParseContext *context,
 			     "<namespace/> name element '%s' doesn't match file name '%s'",
 			     name, ctx->namespace);
 
-	      ctx->current_module = g_ir_module_new (name, version, shared_library, cprefix);
+	      ctx->current_module = _g_ir_module_new (name, version, shared_library, cprefix);
 
 	      ctx->current_module->aliases = ctx->aliases;
 	      ctx->aliases = NULL;
@@ -2789,7 +2909,7 @@ start_element_handler (GMarkupParseContext *context,
 	      ctx->disguised_structures = NULL;
 
 	      for (l = ctx->include_modules; l; l = l->next)
-		g_ir_module_add_include_module (ctx->current_module, l->data);
+		_g_ir_module_add_include_module (ctx->current_module, l->data);
 
 	      g_list_free (ctx->include_modules);
 	      ctx->include_modules = NULL;
@@ -2903,12 +3023,13 @@ start_element_handler (GMarkupParseContext *context,
       break;
     }
 
-  if (ctx->state != STATE_PASSTHROUGH)
+  if (*error == NULL && ctx->state != STATE_PASSTHROUGH)
     {
       g_markup_parse_context_get_position (context, &line_number, &char_number);
       if (!g_str_has_prefix (element_name, "c:"))
-	g_printerr ("%s:%d:%d: warning: dropping to PASSTHROUGH\n",
-		    ctx->file_path, line_number, char_number);
+	g_printerr ("%s:%d:%d: warning: element %s from state %d is unknown, ignoring\n",
+		    ctx->file_path, line_number, char_number, element_name,
+		    ctx->state);
       state_switch (ctx, STATE_PASSTHROUGH);
       ctx->unknown_depth = 1;
     }
@@ -2954,9 +3075,9 @@ require_one_of_end_elements (GMarkupParseContext *context,
   g_set_error (error,
 	       G_MARKUP_ERROR,
 	       G_MARKUP_ERROR_INVALID_CONTENT,
-	       "Unexpected end tag '%s' on line %d char %d; current state=%d",
+	       "Unexpected end tag '%s' on line %d char %d; current state=%d (prev=%d)",
 	       actual_name,
-	       line_number, char_number, ctx->state);
+	       line_number, char_number, ctx->state, ctx->prev_state);
   return FALSE;
 }
 
@@ -3033,6 +3154,13 @@ end_element_handler (GMarkupParseContext *context,
         }
       break;
 
+    case STATE_C_INCLUDE:
+      if (require_end_element (context, ctx, "c:include", element_name, error))
+	{
+          state_switch (ctx, STATE_REPOSITORY);
+        }
+      break;
+
     case STATE_PACKAGE:
       if (require_end_element (context, ctx, "package", element_name, error))
         {
@@ -3051,6 +3179,8 @@ end_element_handler (GMarkupParseContext *context,
     case STATE_ALIAS:
       if (require_end_element (context, ctx, "alias", element_name, error))
 	{
+	  g_free (ctx->current_alias);
+	  ctx->current_alias = NULL;
 	  state_switch (ctx, STATE_NAMESPACE);
 	}
       break;
@@ -3298,6 +3428,7 @@ end_element_handler (GMarkupParseContext *context,
 
     case STATE_PASSTHROUGH:
       ctx->unknown_depth -= 1;
+      g_assert (ctx->unknown_depth >= 0);
       if (ctx->unknown_depth == 0)
         state_switch (ctx, ctx->prev_state);
       break;
@@ -3325,7 +3456,7 @@ cleanup (GMarkupParseContext *context,
   GList *m;
 
   for (m = ctx->modules; m; m = m->next)
-    g_ir_module_free (m->data);
+    _g_ir_module_free (m->data);
   g_list_free (ctx->modules);
   ctx->modules = NULL;
 
@@ -3333,7 +3464,7 @@ cleanup (GMarkupParseContext *context,
 }
 
 /**
- * g_ir_parser_parse_string:
+ * _g_ir_parser_parse_string:
  * @parser: a #GIrParser
  * @namespace: the namespace of the string
  * @filename: (allow-none): Path to parsed file, or %NULL
@@ -3344,16 +3475,15 @@ cleanup (GMarkupParseContext *context,
  * Parse a string that holds a complete GIR XML file, and return a list of a
  * a #GirModule for each &lt;namespace/&gt; element within the file.
  *
- * Returns: (transfer container): a newly allocated list of #GIrModule. The modules themselves
- *  are owned by the #GIrParser and will be freed along with the parser.
+ * Returns: (transfer none): a new #GirModule
  */
-GList *
-g_ir_parser_parse_string (GIrParser           *parser,
-			  const gchar         *namespace,
-			  const gchar         *filename,
-			  const gchar         *buffer,
-			  gssize               length,
-			  GError             **error)
+GIrModule *
+_g_ir_parser_parse_string (GIrParser           *parser,
+			   const gchar         *namespace,
+			   const gchar         *filename,
+			   const gchar         *buffer,
+			   gssize               length,
+			   GError             **error)
 {
   ParseContext ctx = { 0 };
   GMarkupParseContext *context;
@@ -3379,6 +3509,7 @@ g_ir_parser_parse_string (GIrParser           *parser,
 
   g_markup_parse_context_free (context);
 
+  ctx.state = STATE_START;
   context = g_markup_parse_context_new (&markup_parser, 0, &ctx, NULL);
   if (!g_markup_parse_context_parse (context, buffer, length, error))
     goto out;
@@ -3405,11 +3536,13 @@ g_ir_parser_parse_string (GIrParser           *parser,
 
   g_markup_parse_context_free (context);
 
-  return ctx.modules;
+  if (ctx.modules)
+    return ctx.modules->data;
+  return NULL;
 }
 
 /**
- * g_ir_parser_parse_file:
+ * _g_ir_parser_parse_file:
  * @parser: a #GIrParser
  * @filename: filename to parse
  * @error: return location for a #GError, or %NULL
@@ -3420,14 +3553,14 @@ g_ir_parser_parse_string (GIrParser           *parser,
  * Returns: (transfer container): a newly allocated list of #GIrModule. The modules themselves
  *  are owned by the #GIrParser and will be freed along with the parser.
  */
-GList *
-g_ir_parser_parse_file (GIrParser   *parser,
-			const gchar *filename,
-			GError     **error)
+GIrModule *
+_g_ir_parser_parse_file (GIrParser   *parser,
+			 const gchar *filename,
+			 GError     **error)
 {
   gchar *buffer;
   gsize length;
-  GList *modules;
+  GIrModule *module;
   const char *slash;
   char *dash;
   char *namespace;
@@ -3458,13 +3591,13 @@ g_ir_parser_parse_file (GIrParser   *parser,
   if (!g_file_get_contents (filename, &buffer, &length, error))
     return NULL;
 
-  modules = g_ir_parser_parse_string (parser, namespace, filename, buffer, length, error);
+  module = _g_ir_parser_parse_string (parser, namespace, filename, buffer, length, error);
 
   g_free (namespace);
 
   g_free (buffer);
 
-  return modules;
+  return module;
 }
 
 
