@@ -27,6 +27,7 @@ from xml.etree.cElementTree import parse
 
 from . import ast
 from . import message
+from . import utils
 from .transformer import TransformerException
 
 # GParamFlags
@@ -136,6 +137,7 @@ class GDumpParser(object):
             variant = self._namespace.get('Variant')
             assert variant is not None
             variant.add_gtype('GVariant', 'g_variant_get_gtype')
+            variant.c_symbol_prefix = 'variant'
             # Work around scanner bug
             variant.disguised = False
 
@@ -166,7 +168,8 @@ blob containing data gleaned from GObject's primitive introspection."""
                 raise SystemExit(e)
             return parse(out_path)
         finally:
-            shutil.rmtree(self._binary.tmpdir)
+            if not utils.have_debug_flag('save-temps'):
+                shutil.rmtree(self._binary.tmpdir)
 
     def _create_gobject(self, node):
         type_name = 'G' + node.name
@@ -232,6 +235,7 @@ blob containing data gleaned from GObject's primitive introspection."""
             self._create_gobject(record)
         elif record.name == 'InitiallyUnownedClass':
             record.fields = self._namespace.get('ObjectClass').fields
+            record.disguised = False
 
     # Introspection over the data we get from the dynamic
     # GObject/GType system out of the binary
@@ -251,13 +255,36 @@ blob containing data gleaned from GObject's primitive introspection."""
             raise ValueError("Unhandled introspection XML tag %s", xmlnode.tag)
 
     def _introspect_enum(self, xmlnode):
+        type_name = xmlnode.attrib['name']
+        (get_type, c_symbol_prefix) = self._split_type_and_symbol_prefix(xmlnode)
+        try:
+            enum_name = self._transformer.strip_identifier(type_name)
+        except TransformerException, e:
+            message.fatal(e)
+
+        # The scanned member values are more accurate than the values that the
+        # we dumped from GEnumValue.value because GEnumValue.value has the
+        # values as a 32-bit signed integer, even if they were unsigned
+        # in the source code.
+        previous_values = {}
+        previous = self._namespace.get(enum_name)
+        if isinstance(previous, (ast.Enum, ast.Bitfield)):
+            for member in previous.members:
+                previous_values[member.name] = member.value
+
         members = []
         for member in xmlnode.findall('member'):
             # Keep the name closer to what we'd take from C by default;
             # see http://bugzilla.gnome.org/show_bug.cgi?id=575613
             name = member.attrib['nick'].replace('-', '_')
+
+            if name in previous_values:
+                value = previous_values[name]
+            else:
+                value = member.attrib['value']
+
             members.append(ast.Member(name,
-                                      member.attrib['value'],
+                                      value,
                                       member.attrib['name'],
                                       member.attrib['nick']))
 
@@ -266,12 +293,7 @@ blob containing data gleaned from GObject's primitive introspection."""
             klass = ast.Bitfield
         else:
             klass = ast.Enum
-        type_name = xmlnode.attrib['name']
-        (get_type, c_symbol_prefix) = self._split_type_and_symbol_prefix(xmlnode)
-        try:
-            enum_name = self._transformer.strip_identifier(type_name)
-        except TransformerException, e:
-            message.fatal(e)
+
         node = klass(enum_name, type_name,
                      gtype_name=type_name,
                      c_symbol_prefix=c_symbol_prefix,
@@ -397,9 +419,7 @@ different --identifier-prefix.""" % (xmlnode.attrib['name'], self._namespace.ide
             node.properties.append(ast.Property(
                 pspec.attrib['name'],
                 ast.Type.create_from_gtype_name(ctype),
-                readable, writable, construct, construct_only,
-                ctype,
-                ))
+                readable, writable, construct, construct_only))
         node.properties = node.properties
 
     def _introspect_signals(self, node, xmlnode):
